
/******************************************************************************
* MODULE     : edit_search.gen.cc
* DESCRIPTION: search and query replace
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#module code_edit_search

/******************************************************************************
* Test whether we found a match
******************************************************************************/

path
edit_interface::test (path p, tree t) {
  if (N(t) != 0) return p;
  tree st= subtree (et, path_up (p));
  if (N(st) != 0) return p;
  int l= last_item (p);
  // cout << "Test with " << st->label << " at " << l << "\n";
  if (N(st->label) < (N(t->label) + l)) return p;
  if (st->label (l, l + N(t->label)) != t->label) return p;
  string mode= get_env_value (MODE, p)->label;
  string lan = get_env_value (LANGUAGE (mode), p)->label;
  if (search_mode != mode) return p;
  if (search_lan != lan) return p;
  return path_add (p, N(t->label));
}

/******************************************************************************
* Traversal of the edit tree
******************************************************************************/

void
edit_interface::step_ascend (bool forward) {
  // cout << "Step ascend at " << search_at << "\n";
  search_at= path_add (path_up (search_at), forward? 1: -1);
  tree st  = subtree (et, path_up (search_at));
  int  l   = last_item (search_at);

  if (forward) {
    if (l == N(st)) {
      if (atom (search_at)) search_at= path ();
      else step_ascend (forward);
    }
    else step_descend (forward);
  }
  else {
    if (l == -1) {
      if (atom (search_at)) search_at= path ();
      else step_ascend (forward);
    }
    else step_descend (forward);
  }
}

void
edit_interface::step_descend (bool forward) {
  // cout << "Step descend at " << search_at << "\n";
  tree st= subtree (et, search_at);
  int last= (N(st)==0? N(st->label): N(st)-1);
  search_at= path_down (search_at, forward? 0: last);
  if (N(st) != 0)
    if (is_concat (st) || is_document (st))
      step_descend (forward);
}

void
edit_interface::step_horizontal (bool forward) {
  // cout << "Step horizontal at " << search_at << "\n";
  if (nil (search_at)) step_descend (forward);
  else {
    tree st  = subtree (et, path_up (search_at));
    int  l   = last_item (search_at);

    if (forward) {
      if (l == right_index (st)) step_ascend (forward);
      else {
	if ((N(st)==0) && (st->label[l]=='<')) {
	  string s= st->label;
	  while ((l<N(s)) && (s[l]!='>')) l++;
	  if (l<N(s)) l++;
	  search_at= path_down (path_up (search_at), l);
	}
	else search_at= path_inc (search_at);
      }
    }
    else {
      if (l == 0) step_ascend (forward);
      else {
	if ((N(st)==0) && (st->label[l-1]=='>')) {
	  string s= st->label;
	  l--;
	  while ((l>0) && (s[l]!='<')) l--;
	  search_at= path_down (path_up (search_at), l);
	}
	else search_at= path_dec (search_at);
      }
    }
  }
}

void
edit_interface::next_match (bool forward) {
  // cout << "Next match at " << search_at << "\n";
  while (TRUE) {
    if (nil (search_at)) {
      start_p= copy (tp);
      end_p  = copy (tp);
      notify_change (THE_SELECTION);
      return;
    }
    search_end= test (search_at, search_what);
    if (search_end != search_at) {
      start_p= copy (search_at);
      end_p  = copy (search_end);
      notify_change (THE_SELECTION);
      set_cursor (copy (search_at));
      return;
    }
    step_horizontal (forward);
  }
}

/******************************************************************************
* Searching
******************************************************************************/

void
edit_interface::search_start (bool flag) {
  string r ("forward search");
  if (!flag) r= "backward search";

  input_mode  = INPUT_SEARCH;
  forward     = flag;
  search_mode = copy (get_env_value (MODE)->label);
  search_lan  = copy (get_env_value (LANGUAGE (search_mode))->label);
  search_at   = tp;
  search_what = tree ("");
  search_stack= list<path> ();
  keybd_stack = tree ("");
  set_message ("Searching", r);
}

void
edit_interface::search_next (bool forward) {
  string r ("forward search");
  if (!forward) r= "backward search";

  next_match (forward);
  if (nil (search_at)) {
    set_message ("No more matches for#" * search_what->label, r);
    cerr << '\a';
  }
  else set_message ("Searching#" * search_what->label, r);
}

void
edit_interface::search_button_next () {
  input_mode= INPUT_SEARCH;
  search_stack= list<path> (copy (search_at), search_stack);
  keybd_stack = tree (BACKUP, "<menu>", keybd_stack);
  step_horizontal (forward);
  search_next (forward);
}

bool
edit_interface::search_keypress (string s) {
  if (N(s)==1) {
    if (N (search_what) == 0) {
      search_stack= list<path> (copy (search_at), search_stack);
      keybd_stack = tree (BACKUP, copy (s), keybd_stack);
      search_what = tree (search_what->label * s);
      search_next (forward);
    }
  }
  else {
    if ((s == "<left>") || (s == "<right>") ||
	(s == "<up>") || (s == "<down>") ||
	(s == "<page-up>") || (s == "<page-down>") ||
	(s == "<begin>") || (s == "<end>"))
      {
	set_input_normal ();
	return FALSE;
      }
    if ((s == "<ctrl-c>") || (s == "<ctrl-g>"))
      set_input_normal ();
    else if (s == "<ctrl-r>") {
      search_stack= list<path> (copy (search_at), search_stack);
      keybd_stack = tree (BACKUP, copy (s), keybd_stack);
      step_horizontal (FALSE);
      search_next (FALSE);
    }
    else if (s == "<ctrl-s>") {
      search_stack= list<path> (copy (search_at), search_stack);
      keybd_stack = tree (BACKUP, copy (s), keybd_stack);
      step_horizontal (TRUE);
      search_next (TRUE);
    }
    else if ((s == "<del>") || (s == "<bs>")) {
      if (nil (search_stack))
	set_input_normal ();
      else if (atom (search_stack)) {
	set_cursor (search_stack->item);
	set_input_normal ();
      }
      else {
	search_at   = search_stack->item;
	search_stack= search_stack->next;
	if (N(keybd_stack[0]->label)==1)
	  search_what= tree (search_what->label (0, N(search_what->label)-1));
	keybd_stack = keybd_stack[1];
	search_next (forward);
      }
    }
  }
  return TRUE;
}

/******************************************************************************
* Replacing
******************************************************************************/

void
edit_interface::replace_start (tree what, tree by, bool flag) {
  input_mode  = INPUT_REPLACE;
  forward     = flag;
  search_mode = copy (get_env_value (MODE)->label);
  search_lan  = copy (get_env_value (LANGUAGE (search_mode))->label);
  search_at   = tp;
  search_what = copy (what);
  replace_by  = copy (by);
  nr_replaced = 0;
  replace_next ();
}

void
edit_interface::replace_next () {
  string r ("forward replace");
  if (!forward) r= "backward replace";

  next_match (forward);
  if (nil (search_at)) {
    string l= "Replaced#" * as_string (nr_replaced) * "#occurrences";
    if (nr_replaced == 0) l= "No matches found";
    if (nr_replaced == 1) l= "Replaced one occurrence";
    set_message (l, r);
    cerr << '\a';
    set_input_normal ();
  }
  else set_message ("Replace (y,n,a)?", r);
}

bool
edit_interface::replace_keypress (string s) {
  if ((s == "<ctrl-c>") || (s == "<ctrl-g>") || (s == "<esc>"))
    set_input_normal ();
  else if (s == "y") {
    nr_replaced++;
    set_cursor (copy (search_end));
    cut (search_at, search_end);
    insert (copy (replace_by));
    search_at= copy (tp);
    replace_next ();
  }
  else if (s == "n") {
    step_horizontal (forward);
    replace_next ();
  }
  else if (s == "a") {
    while (!nil (search_at)) {
      nr_replaced++;
      set_cursor (copy (search_end));
      cut (search_at, search_end);
      insert (copy (replace_by));
      search_at= copy (tp);
      replace_next ();
    }
  }
  return TRUE;
}

#endmodule // code_edit_search
