
/******************************************************************************
* MODULE     : modify_session.gen.cc
* DESCRIPTION: input and output handling for programming sessions
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <list.gen.h>
#include <connect.gen.h>

#module code_modify_session
#import list (path)
#import connect

/******************************************************************************
* Compute the context of a path in the edit tree
******************************************************************************/

void
compute_context (tree t, path where, list<path>& l, list<path>& r) {
  list<path> e;
  l= list<path>();
  r= list<path>();
  path p= start_forward (t);
  while (!nil (p)) {
    tree u= subtree (t, p);
    if (is_func (u, BEGIN)) e= list<path> (p, e);
    if (is_func (u, END))
      if (!nil (e)) {
	if ((path_inf (e->item, where) && path_inf (where, p)) ||
	    (where == path_down (e->item, 1)) || (where == path_down (p, 0)))
	  {
	    l= list<path> (e->item, l);
	    r= list<path> (p, r);
	  }
	e= e->next;
      }
    p= increment (t, p);
  }
}

string
compute_context (tree t, path where) {
  string s;
  list<path> l, r;
  compute_context (t, where, l, r);
  while (!nil (l)) {
    tree u= subtree (t, l->item);
    s << (u[0]->label * "#");
    l= l->next;
  }
  return s;
}

/******************************************************************************
* Routines on environments
******************************************************************************/

bool
edit_modify::inside_environment (string name) {
  list<path> l, r;
  compute_context (et, tp, l, r);
  while ((!nil (l)) && (!nil (r))) {
    tree u= subtree (et, l->item);
    if (u == tree (BEGIN, name)) return TRUE;
    l= l->next; r= r->next;
  }
  return FALSE;
}

void
edit_modify::goto_end_environment (string name) {
  list<path> l, r;
  compute_context (et, tp, l, r);
  while ((!nil (l)) && (!nil (r))) {
    tree u= subtree (et, l->item);
    if (u == tree (BEGIN, name)) {
      set_cursor (path_down (r->item, 1));
      break;
    }
    l= l->next; r= r->next;
  }
}

tree
edit_modify::interior_environment (string name) {
  list<path> l, r;
  compute_context (et, tp, l, r);
  while ((!nil (l)) && (!nil (r))) {
    tree u= subtree (et, l->item);
    if (u == tree (BEGIN, name)) {
      path p1= path_down (l->item, 1);
      path p2= path_down (r->item, 0);
      tree back= compute_selection (et, p1, p2);
      // cout << "Interior : " << back << "\n";
      return back;
    }
    l= l->next; r= r->next;
  }
  return "";
}

void
edit_modify::cut_environment (string name) {
  list<path> l, r;
  compute_context (et, tp, l, r);
  while ((!nil (l)) && (!nil (r))) {
    tree u= subtree (et, l->item);
    if (u == tree (BEGIN, name)) {
      path p1= path_down (l->item, 0);
      path p2= path_down (r->item, 1);
      cut (p1, p2);
      return;
    }
    l= l->next; r= r->next;
  }
}

/******************************************************************************
* Other useful subroutines
******************************************************************************/

static bool
is_empty (tree t) {
  if (N(t) == 0) return (t == "");
  if (is_document (t) || is_concat (t)) {
    int i;
    for (i=0; i<N(t); i++)
      if (!is_empty (t[i])) return FALSE;
    return TRUE;
  }
  return FALSE;
}

/******************************************************************************
* Routines for sessions
******************************************************************************/

void
edit_modify::make_session (string lan, string session) {
  make_active_begin (tree ("session", lan, session));
  if (lan == "shell") start_input ();
  else if (connection_declared (lan)) connect ();
  else start_input ();
}

void
edit_modify::start_input () {
  if (inside_environment ("input")) {
    goto_end_environment ("input");
    insert_return ();
  }
  else if (inside_environment ("output")) {
    if (is_empty (interior_environment ("output")))
      cut_environment ("output");
    else {
      goto_end_environment ("output");
      path p1= path_up (tp);
      path p2= path_up (p1);
      if (is_document (subtree (et, p2)))
	remove_return (path_dec (p1));
      else if (is_concat (subtree (et, p2)) &&
	       (last_item (p1) == 0) &&
	       is_document (subtree (et, path_up (p2))))
	  remove_return (path_dec (p2));
      insert_return ();
    }
  }
  tree bi (BEGIN, "input"), ei (END, "input");
  insert_tree (tree (CONCAT, bi, ei), path (0, 1));
}

void
edit_modify::start_output () {
  if (inside_environment ("input")) {
    goto_end_environment ("input");
    insert_return ();
  }
  else if (inside_environment ("output")) {
    goto_end_environment ("output");
    insert_return ();
  }
  tree bi (BEGIN, "output"), ei (END, "output");
  insert_tree (tree (CONCAT, bi, ei), path (0, 1));
}

#endmodule // code_modify_session
